/*
* Copyright (c) 2006, Berges Allmenndigitale Rdgivningstjeneste
* All rights reserved.
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of Berges Allmenndigitale Rdgivningstjeneste nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE BERGES AND CONTRIBUTORS ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL THE BERGES AND CONTRIBUTORS BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#ifdef _WIN32
# include <windows.h>
# include <io.h>
# include <direct.h>
static int optind=1;
#else
# define _GNU_SOURCE
# include <getopt.h>
# include <pthread.h>
# include <unistd.h>
# include <dlfcn.h>
typedef char TCHAR;
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "../js/src/jsapi.h"
#include "../js/src/jsprf.h"
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include "version.h"

#ifdef UNICODE
#define JS_GetStringTChars JS_GetStringChars
#define JS_NewStringTCopyZ JS_NewUCStringCopyZ
#define JS_DefineProperty(cx, obj, name, value, getter, setter, attrs) \
  JS_DefineUCProperty((cx),(obj),(name),wcslen(name),(value),(getter),(setter),(attrs))
#else
#define JS_NewStringTCopyZ JS_NewStringCopyZ
#define JS_GetStringTChars JS_GetStringBytes
#endif

static JSObject *glob;
static char **argv;
static char **envp;
static int argc;
static JSRuntime *rt;

#ifdef _WIN32
static HINSTANCE hInstance=0;  // handle to current instance
static HINSTANCE hPrevInstance=0;  // handle to previous instance
static LPSTR lpCmdLine=0;      // pointer to command line
static int nCmdShow=0;          // show state of window

__declspec(dllexport)
HINSTANCE JSEXT_get_hInstance(void) {
	return hInstance;
}
__declspec(dllexport)
HINSTANCE JSEXT_get_hPrevInstance(void) {
	return hPrevInstance;
}
__declspec(dllexport)
LPSTR JSEXT_get_lpCmdLine(void) {
	return lpCmdLine;
}
__declspec(dllexport)
int JSEXT_get_nCmdShow(void) {
	return nCmdShow;
}

#endif

static void printhelp() {
#ifdef _WIN32
	puts("jsext [OPTION]... [FILE] [ARGUMENTS]\n\n"
       "Interprets FILE as a JavaScript program\n\n"
       "Options:\n"
       "/r RCFILE\n"
       "\tExecute commands from RCFILE instead of the\n"
	   "\tfile named jsext.rc in the same directory as\n"
	   "\tthe executable file.\n"
       "/n\n"
       "\tDo not read and execute the initialization file\n"
	   "\tnamed jsext.rc in the same directory as the\n"
       "\texecutable file.\n"
       "/e EXPR\n"
       "\tEvaluate the expression\n"
       "/?\n"
       "\tPrint help\n");
#else
	puts("jsext [OPTION]... [FILE] [ARGUMENTS]\n\n"
       "Interprets FILE as a JavaScript program\n\n"
       "Options:\n"
       "-r, --rcfile=RCFILE\n"
       "\tExecute commands from RCFILE instead of the system wide\n"
	   "\tinitialization file" sysconfdir "/jsext.jsextrc-" jsext_version
	   "\tand the standard personal initialization file\n"
	   "\t~/.jsextrc-" jsext_version "\n"
       "-n, --norc\n"
       "\tDo not read and execute the system wide initialization file\n"
       "\t" sysconfdir "/jsext.jsextrc-" jsext_version "\n"
       "\tand the personal initialization file ~/.jsextrc-" jsext_version "\n"
       "-e, --eval=EXPR\n"
       "\tEvaluate the expression\n"
       "-h, --help\n"
       "\tPrint help\n");
#endif
}

JSBool JSCC_ReportException(JSContext *cx, char *format, ...) {
  int len;
  char *msg;
  JSString *Str;
  va_list va;
  jsval str;

  va_start(va, format);
#ifdef _WIN32
  msg=JS_malloc(cx, 81);
  va_start(va, format);
  len=_vsnprintf(msg,80,format,va);
  msg[80]=0;
#else
  len=vsnprintf(NULL,0,format,va);
  va_end(va);
  msg=JS_malloc(cx, len+1);
  va_start(va, format);
  vsprintf(msg,format,va);
#endif
  va_end(va);
  Str=JS_NewString(cx, msg, len);
  str=STRING_TO_JSVAL(Str);
  JS_SetPendingException(cx, str);

  return JS_FALSE;
}

static JSBool
env_setProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
/* XXX porting may be easy, but these don't seem to supply setenv by default */
#if !defined XP_BEOS && !defined XP_OS2 && !defined SOLARIS
    JSString *idstr, *valstr;
    const TCHAR *name, *value;
    int rv;

    idstr = JS_ValueToString(cx, id);
    valstr = JS_ValueToString(cx, *vp);
    if (!idstr || !valstr)
        return JS_FALSE;
    name = JS_GetStringTChars(idstr);
    value = JS_GetStringTChars(valstr);
#ifdef XP_WIN
	if (SetEnvironmentVariable(name,value)) rv=0;
	else rv=-1;
#elif defined HPUX || defined OSF1 || defined IRIX
    {
        char *waste = JS_smprintf("%s=%s", name, value);
        if (!waste) {
            JS_ReportOutOfMemory(cx);
            return JS_FALSE;
        }

        rv = putenv(waste);
        /*
         * HPUX9 at least still has the bad old non-copying putenv.
         *
         * Per mail from <s.shanmuganathan@digital.com>, OSF1 also has a putenv
         * that will crash if you pass it an auto char array (so it must place
         * its argument directly in the char *environ[] array).
         */
//        free(waste);
    }
#else
    rv = setenv(name, value, 1);
#endif
    if (rv < 0) {
        JSCC_ReportException(cx, "can't set envariable %s to %s", name, value);
        return JS_FALSE;
    }
    *vp = STRING_TO_JSVAL(valstr);
#endif /* !defined XP_BEOS && !defined XP_OS2 && !defined SOLARIS */
    return JS_TRUE;
}

static JSBool
env_enumerate(JSContext *cx, JSObject *obj)
{
    static JSBool reflected;
    TCHAR **evp, *name, *value, *es, **origevp;
	int nenv=0;
	int envcap=0;
    JSString *valstr;
    JSBool ok;

    if (reflected)
        return JS_TRUE;

#ifdef _WIN32
	es= name = (TCHAR *)GetEnvironmentStrings();
	envcap=128;
	origevp=(TCHAR **)malloc(sizeof(TCHAR *)*envcap);
	for(;;) {
		if (nenv==envcap) {
			envcap*=2;
			origevp=(TCHAR **)realloc(origevp, sizeof(TCHAR *)*envcap);
		}
		if (*name) {
			origevp[nenv++]=name;
			while (*name) name++;
			name++;
		} else {
			origevp[nenv]=0;
			break;
		}
	};
	evp=origevp;
#else
    evp = (TCHAR **)JS_GetPrivate(cx, obj);
#endif
    for (; (name = *evp) != NULL; evp++) {
		value=name;
		while (*value && *value!='=') value++;
        if (!*value)
            continue;
        *value++ = '\0';
        valstr = JS_NewStringTCopyZ(cx, value);
        if (!valstr) {
            ok = JS_FALSE;
        } else {
            ok = JS_DefineProperty(cx, obj, name, STRING_TO_JSVAL(valstr),
                                   NULL, NULL, JSPROP_ENUMERATE);
        }
        value[-1] = '=';
        if (!ok) goto failure;
    }

    reflected = JS_TRUE;

#ifdef _WIN32
	free(origevp);
	FreeEnvironmentStrings(es);
#endif

	return JS_TRUE;
failure:
#ifdef _WIN32
	free(origevp);
	FreeEnvironmentStrings(es);
#endif
	return JS_FALSE;
}

static JSBool
env_resolve(JSContext *cx, JSObject *obj, jsval id, uintN flags,
            JSObject **objp)
{
    int vallen;
	JSString *idstr, *valstr;
    TCHAR *name, *value=0;

    if (flags & JSRESOLVE_ASSIGNING)
        return JS_TRUE;

    idstr = JS_ValueToString(cx, id);
    if (!idstr)
        return JS_FALSE;
    name = JS_GetStringTChars(idstr);
#ifdef _WIN32
	vallen=GetEnvironmentVariable(name,0,0);
	if (vallen) {
		value=malloc(vallen*sizeof(TCHAR));
		GetEnvironmentVariable(name,value,vallen);
	}
#else
    value = getenv(name);
#endif
    if (value) {
        valstr = JS_NewStringTCopyZ(cx, value);
        if (!valstr)
			goto failure;
        if (!JS_DefineProperty(cx, obj, name, STRING_TO_JSVAL(valstr),
                               NULL, NULL, JSPROP_ENUMERATE)) {
            goto failure;
        }
        *objp = obj;
    }

#ifdef _WIN32
	if (value) free(value);
#endif
    return JS_TRUE;

failure:
#ifdef _WIN32
	if (value) free(value);
#endif
	return JS_FALSE;
}

static JSClass env_class = {
    "environment",
    JSCLASS_HAS_PRIVATE | JSCLASS_NEW_RESOLVE,
    JS_PropertyStub,
    JS_PropertyStub,
    JS_PropertyStub,
    env_setProperty,
    env_enumerate,
    (JSResolveOp) env_resolve,
    JS_ConvertStub,
    JS_FinalizeStub
};

static JSBool JS_DLL_CALLBACK JSEXT_load(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval) {
  char *filename;
  int fd;
  struct stat S;
  char *buf;
  int shebanglen=0;
  JSScript *script;

  if (argc!=1) {
    JSCC_ReportException(cx, "Wrong number of arguments");
    return JS_FALSE;
  }

  if (!JS_ConvertArguments(cx, argc, argv, "s", &filename)) {
    return JS_FALSE;
  }

#ifdef _WIN32
  fd=_open(filename,O_RDONLY | O_BINARY);
#else
  fd=open(filename,O_RDONLY);
#endif
  if (fd==-1) {
    // File does not exist, throw exception
    JSCC_ReportException(cx, "load: Could not open %s",filename);
    return JS_FALSE;
  }
  fstat(fd, &S);
  buf=malloc(S.st_size);
  if (!buf) {
    JSCC_ReportException(cx, "load: Out of memory for %s", filename);
    goto failure;
  }
#ifdef _WIN32
  if (_read(fd, buf, S.st_size)!=S.st_size) {
#else
  if (read(fd, buf, S.st_size)!=S.st_size) {
#endif
    JSCC_ReportException(cx, "load: Error loading %s",filename);
    goto failure;
  }

  if (S.st_size>1 && buf[0]=='#' && buf[1]=='!') {
    // shebang
    while (buf[shebanglen]!='\n' &&
	   buf[shebanglen]!='\r') shebanglen++;
    shebanglen++;
    if (shebanglen<S.st_size &&
	(buf[shebanglen]=='\n' ||
	 buf[shebanglen]=='\r') &&
	buf[shebanglen]!=buf[shebanglen-1]) shebanglen++;
  }

  script=JS_CompileScript(cx, obj, buf+shebanglen, S.st_size-shebanglen, filename, shebanglen?2:1);
  if (!script) goto failure;
  
  free(buf);
  buf=0;
  
  *rval=JSVAL_ZERO;

#ifdef _WIN32
  _close(fd);
#else
  close(fd);
#endif
  fd=0;
  
  if (!JS_ExecuteScript(cx, obj, script, rval)) goto failure;
  return JS_TRUE;

 failure:
  if (buf) free(buf);
  if (fd) close(fd);
  return JS_FALSE;
}

#ifdef _WIN32
#define DLLACCESS __declspec(dllexport)
#else
#define DLLACCESS
#endif

static JSBool dl_new(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval);
void JS_DLL_CALLBACK JSEXT_dl_finalize(JSContext *cx, JSObject *obj);

static JSClass JSEXT_dl_class = {
    "dl",
    JSCLASS_HAS_PRIVATE,
    JS_PropertyStub,
    JS_PropertyStub,
    JS_PropertyStub,
    JS_PropertyStub,
    JS_EnumerateStub,
    JS_ResolveStub,
    JS_ConvertStub,
    JSEXT_dl_finalize
};

static JSObject *JSEXT_dl_proto;

#ifdef UNICODE
JSBool DLLACCESS
JSEXT_dl_new(JSContext *cx, JSObject *obj, TCHAR *filename, jsval *rval);

JSBool DLLACCESS
JSEXT_dl_newA(JSContext *cx, JSObject *obj, char *filename, jsval *rval) {
	TCHAR *UCfilename=(TCHAR *)malloc((strlen(filename)+1)*2);
	TCHAR *p=UCfilename;
	JSBool ret;
	while (*filename) *(p++)=*(filename++);
	*p=0;
	ret=JSEXT_dl_new(cx, obj, UCfilename, rval);
	free(UCfilename);
	return ret;
}

JSBool DLLACCESS
JSEXT_dl_new(JSContext *cx, JSObject *obj, TCHAR *filename, jsval *rval) {
#else
#define JSEXT_dl_new JSEXT_dl_newA 
JSBool DLLACCESS
JSEXT_dl_newA(JSContext *cx, JSObject *obj, TCHAR *filename, jsval *rval) {
#endif
  JSObject *object;
  JSNative JSEXT_init=0;
  JSString *JSfilename;
  jsval tmpval;

  if (JS_FALSE==JS_InstanceOf( cx, obj, &JSEXT_dl_class, 0)) { // not called with new
    object=JS_NewObject( cx, &JSEXT_dl_class, JSEXT_dl_proto, 0);
    *rval=OBJECT_TO_JSVAL(object);
  } else {
    object=obj;
  }

#ifdef _WIN32
  {
  HMODULE handle;
    
  if (filename)
	handle=LoadLibraryEx(filename, NULL, 0);
  else
	handle=GetModuleHandle(NULL);

  JS_SetPrivate(cx, object, (void *)handle);

  if (!handle) {
    return JS_FALSE;
  }
  if (filename)
	  JSEXT_init=(JSNative) GetProcAddress(handle,"JSEXT_init");
  }
#else
  void *dl;
    
  if (filename)
    dl=dlopen(filename, RTLD_LAZY | RTLD_GLOBAL);
  else
	dl=dlopen(0,RTLD_LAZY);

  JS_SetPrivate(cx, object, dl);

  if (!dl) {
    return JS_FALSE;
  }

  if (filename)
	  JSEXT_init=(JSNative) dlsym(dl,"JSEXT_init");
#endif

  if (JSEXT_init) { // Call init with filename as only argument
    JSString *tmpstr;
    jsval argv[1];
    jsval rval=JSVAL_VOID;

    tmpstr=JS_NewStringTCopyZ(cx, filename);
    JS_AddRoot(cx, &tmpstr);
    argv[0]=STRING_TO_JSVAL(tmpstr);

    (*JSEXT_init)(cx, obj, 1, argv, &rval);

    JS_RemoveRoot(cx, &tmpstr);
  }

  if (filename)
	JSfilename=JS_NewStringTCopyZ(cx, filename);
  else
#ifdef UNICODE
	JSfilename=JS_NewStringTCopyZ(cx, L"main program");
#else
	JSfilename=JS_NewStringTCopyZ(cx, "main program");
#endif

  tmpval=STRING_TO_JSVAL(JSfilename);
  JS_SetProperty(cx, object, "filename", &tmpval);

  return JS_TRUE;
}

static JSBool
dl_new(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
  TCHAR *filename;
  JSBool res;

  if (argc==1 && JSVAL_IS_NULL(argv[0]))
	filename=0;
#ifdef UNICODE
  else if (!JS_ConvertArguments(cx, argc, argv, "W", &filename)) {
#else
  else if (!JS_ConvertArguments(cx, argc, argv, "s", &filename)) {
#endif
    JSCC_ReportException(cx, "Wrong parameter for dl");
    return JS_FALSE;
  }
  
  res=JSEXT_dl_new(cx, obj, filename, rval);
  if (res==JS_FALSE) {
#ifdef _WIN32
    JSCC_ReportException(cx, "can't open dl '%s'. Error code=%d", filename, GetLastError());
#else
    JSCC_ReportException(cx, "can't open dl %s", dlerror());
#endif
    return JS_FALSE;
  }

  return JS_TRUE;
}

static JSBool JS_DLL_CALLBACK JSX_dl_symbolExists(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval) {
  char *symbol;

  if (argc!=1) {
    JSCC_ReportException(cx, "Wrong number of arguments");
    return JS_FALSE;
  }

  if (!JS_ConvertArguments(cx, argc, argv, "s", &symbol)) {
    return JS_FALSE;
  }

#ifdef _WIN32
  if (GetProcAddress((HMODULE)JS_GetPrivate(cx, obj),symbol)!=0)
	*rval=JSVAL_TRUE;
  else
	  *rval=JSVAL_FALSE;
#else
  if (dlsym((void *)JS_GetPrivate(cx, obj),symbol)!=0)
	*rval=JSVAL_TRUE;
  else
	  *rval=JSVAL_FALSE;
#endif

  return JS_TRUE;
}


void JS_DLL_CALLBACK JSEXT_dl_finalize(JSContext *cx, JSObject *obj) {
#ifdef _WIN32
  HMODULE handle=(HMODULE) JS_GetPrivate(cx, obj);
  if (!handle) return;
//  FreeLibrary(handle);
#else
  void *dl=JS_GetPrivate(cx, obj);
  if (!dl) return;
  //  dlclose(dl);
#endif
}

static JSBool
magic_resolve(JSContext *cx, JSObject *obj, jsval id, uintN flags,
            JSObject **objp)
{
  JSString *idstr;
  const char *name;
  jsval resolveval;
  JSFunction *Func;
  jsval args[2];
  jsval rval;

  idstr = JS_ValueToString(cx, id);
  if (!idstr)
    return JS_FALSE;
  name = JS_GetStringBytes(idstr);
  if (name[0]=='_') {
    *objp=NULL;
    return JS_TRUE; // don't resolve "_resolve".
  }
  
  if (JS_FALSE==JS_GetProperty(cx, obj, "_resolve", &resolveval) ||
      resolveval==JSVAL_VOID) {
    *objp=NULL;
    return JS_TRUE;
  }

  Func=JS_ValueToFunction(cx, resolveval);
  if (!Func) {
    *objp=NULL;
    return JS_TRUE;
  }

  args[0]=id;
  if (flags & JSRESOLVE_ASSIGNING) args[1]=JSVAL_TRUE;
  else args[1]=JSVAL_FALSE;
  
  if (!JS_CallFunction(cx, obj, Func, 2, args, &rval) ||
      rval!=JSVAL_TRUE) {
    *objp=NULL;
    return JS_TRUE;
  }

  *objp=obj;
  return JS_TRUE;
}

static JSBool
magic_enumerate(JSContext *cx, JSObject *obj)
{
  jsval enumval;
  JSFunction *Func;
  jsval rval;

  if (JS_FALSE==JS_GetProperty(cx, obj, "_enumerate", &enumval) ||
      enumval==JSVAL_VOID) {
    return JS_TRUE;
  }

  Func=JS_ValueToFunction(cx, enumval);
  if (!Func) {
    return JS_TRUE;
  }

  if (!JS_CallFunction(cx, obj, Func, 0, &rval, &rval)) {
    return JS_TRUE;
  }
  
  return JS_TRUE;
}

static JSBool
magic_setProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp) {
  JSString *idstr;
  const TCHAR *name;
  jsval propval;
  JSFunction *Func;
  jsval args[2]={id, *vp};
  jsval rval;
	  
  idstr = JS_ValueToString(cx, id);
  if (!idstr)
    return JS_FALSE;
  name = JS_GetStringTChars(idstr);
  if (name[0]=='_') {
    return JS_TRUE; // don't set "_set".
  }

  if (JS_FALSE==JS_GetProperty(cx, obj, "_set", &propval) ||
      propval==JSVAL_VOID) {
    return JS_TRUE;
  }

  Func=JS_ValueToFunction(cx, propval);
  if (!Func) {
    return JS_TRUE;
  }

  if (!JS_CallFunction(cx, obj, Func, 2, args, &rval)) {
    return JS_TRUE;
  }

  if (rval==JSVAL_FALSE) return JS_FALSE;
  return JS_TRUE;
  
}

static JSBool
magic_getProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp) {
  JSString *idstr;
  const TCHAR *name;
  jsval propval;
  JSFunction *Func;
  jsval args[2]={id, *vp};
  
  idstr = JS_ValueToString(cx, id);
  if (!idstr)
    return JS_FALSE;
  name = JS_GetStringTChars(idstr);
  if (name[0]=='_') {
    return JS_TRUE; // don't get "_get".
  }

  if (JS_FALSE==JS_GetProperty(cx, obj, "_get", &propval) ||
      propval==JSVAL_VOID) {
    return JS_TRUE;
  }

  Func=JS_ValueToFunction(cx, propval);
  if (!Func) {
    return JS_TRUE;
  }

  if (!JS_CallFunction(cx, obj, Func, 2, args, vp)) {
    return JS_TRUE;
  }

  return JS_TRUE;
  
}

JSClass magic_class = {
    "Magic",
    JSCLASS_NEW_RESOLVE,
    JS_PropertyStub,
    JS_PropertyStub,
    magic_getProperty,
    magic_setProperty,
    magic_enumerate,
    (JSResolveOp)  magic_resolve,
    JS_ConvertStub,
    JS_FinalizeStub
};

#ifdef _WINDOWS
static void
my_ErrorReporter(JSContext *cx, const char *message, JSErrorReport *report)
{
  TCHAR errbuf[1024];
  errbuf[0]=0;
#ifdef UNICODE
  if (report->lineno && report->filename) _snwprintf(errbuf, 1024,L"Line %d in %S:",report->lineno,report->filename);
  if (message) _snwprintf(errbuf+wcslen(errbuf),1024-wcslen(errbuf),L"%S\n",message);
  MessageBox(0,errbuf,L"Error",MB_OK);
#else
  if (report->lineno && report->filename) _snprintf(errbuf, 1024,"Line %d in %s:",report->lineno,report->filename);
  if (message) _snprintf(errbuf+strlen(errbuf),1024-strlen(errbuf),"%s\n",message);
  MessageBox(0,errbuf,"Error",MB_OK);
#endif
}
#else
static void
my_ErrorReporter(JSContext *cx, const char *message, JSErrorReport *report)
{
  if (report->lineno && report->filename) fprintf(stderr,"Line %d in %s:",report->lineno,report->filename);
  if (message) fprintf(stderr,"%s\n",message);
}
#endif

static JSBool maybeexec(JSContext *cx, JSObject *obj, char *filename) {

#ifdef _WIN32
  int fd=open(filename,O_RDONLY | O_BINARY);
#else
  int fd=open(filename,O_RDONLY);
#endif
  char *buf;
  JSScript *script;
  struct stat S;
  jsval rval=JSVAL_ZERO;

  if (fd==-1) {
    return JS_TRUE; // File does not exist, but that is okay
  }
  fstat(fd, &S);
  buf=malloc(S.st_size);
  if (!buf) {
    JSCC_ReportException(cx, "Out of memory for %s", filename);
    goto failure;
  }
  if (read(fd, buf, S.st_size)!=S.st_size) {
    JSCC_ReportException(cx, "Error loading %s",filename);
    goto failure;
  }

  script=JS_CompileScript(cx, obj, buf, S.st_size, filename, 1);
  if (!script) goto failure;
  
  free(buf);
  buf=0;
  
  close(fd);
  fd=0;
  
  if (!JS_ExecuteScript(cx, obj, script, &rval)) goto failure;
  return JS_TRUE;

 failure:
  if (buf) free(buf);
  if (fd) close(fd);
  return JS_FALSE;
}

static char *rcfile=0;
static int norc=0;

static JSBool JS_DLL_CALLBACK magic_new(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval) {
  return JS_TRUE;
}

#ifdef _WIN32
__declspec(dllexport)
#endif
JSContext *JSEXT_InitContext(JSRuntime *rt) {
  JSContext *cx=JS_NewContext(rt, 4096);
  JSString *version;
  jsval versionval;
  JSObject *envobj;
  JSObject *args;
  jsval argval;
  int j;

  if (!cx) goto failure;
  JS_SetErrorReporter(cx, my_ErrorReporter);

  glob=JS_NewObject(cx, &magic_class, NULL, NULL);
  if (!glob)
    goto failure;
  if (!JS_InitStandardClasses(cx, glob))
    goto failure;
  if (!JS_InitClass(cx, glob, NULL, &magic_class, magic_new, 0, NULL, NULL, NULL, NULL)) goto failure;

  JSEXT_dl_proto=JS_InitClass(cx, glob, NULL, &JSEXT_dl_class, dl_new, 1, NULL, NULL, NULL, NULL);
  if (JSEXT_dl_proto==0) goto failure;
  JS_SetPrivate(cx, JSEXT_dl_proto, NULL);

  
  if (JS_FALSE ==
      JS_DefineFunction (cx, glob, "load", JSEXT_load, 1, 0))
    goto failure;
  if (JS_FALSE ==
      JS_DefineFunction (cx, JSEXT_dl_proto, "symbolExists", JSX_dl_symbolExists, 1, 0))
    goto failure;

  version = JS_NewStringCopyZ(cx,jsext_version);
  versionval=STRING_TO_JSVAL(version);
  JS_SetProperty (cx, glob, "_version", &versionval);

#ifdef _WIN32
  argval=INT_TO_JSVAL(1);
  JS_SetProperty (cx, glob, "_WIN32", &argval);
#endif
#ifdef _WINDOWS
  argval=INT_TO_JSVAL(1);
  JS_SetProperty (cx, glob, "_WINDOWS", &argval);
#endif

  envobj = JS_DefineObject(cx, glob, "environment", &env_class, NULL, 0);
  if (!envobj || !JS_SetPrivate(cx, envobj, envp))
    goto failure;


  args=JS_NewArrayObject(cx, 0, 0);
  argval=OBJECT_TO_JSVAL(args);
  JS_SetProperty(cx, glob, "arguments", &argval);
  for (j=optind; j<argc; j++) {
    JSString *str=JS_NewStringCopyZ(cx,argv[j]);
    jsval elem=STRING_TO_JSVAL(str);
      JS_SetElement(cx, args, j-optind, &elem);
  }

  if (rcfile && !norc) {
    // Source argument-specifier init file if it exists
    if (JS_FALSE==maybeexec(cx,glob,rcfile)) goto failure;
    
  } else if (!norc) {
#ifdef _WIN32
	// Source init file from same dir as exe file if it exists
    char rc[_MAX_PATH];
	char *lastslash=0;
	if (argv[0][0]=='\\' || (
		strlen(argv[0])>2 && argv[0][1]==':' && argv[0][2]=='\\')) {
		// absolute path to exe file
		strcpy(rc, argv[0]);
	} else {
		// relative path to exe file
		if (strlen(argv[0])>2 && argv[0][1]==':')
			_getdcwd(argv[0][0] & 0x1f, rc, _MAX_PATH);
		else
			_getcwd(rc, _MAX_PATH);
		if (rc[strlen(rc)-1]!='\\' && strlen(rc)<_MAX_PATH-1) strcat(rc,"\\");
		strncat(rc, argv[0], _MAX_PATH-1-strlen(rc));
	}
	// remove name of exe file
	lastslash=strrchr(rc,'\\');
	if (!lastslash) goto failure;
	lastslash[0]=0;
	SetEnvironmentVariableA("JSEXT_EXE_DIR",rc);
	strcat(rc,"\\jsextrc");
    if (JS_FALSE==maybeexec(cx,glob,rc)) goto failure;

#else
    char rc[1000];
    // Source system-wide init file if it exists
    sprintf(rc,sysconfdir "/jsext.jsextrc-" jsext_version);
    if (JS_FALSE==maybeexec(cx,glob,rc)) goto failure;
    // Source user init file if it exists
    sprintf(rc,"%s/.jsextrc-" jsext_version,getenv("HOME"));
    if (JS_FALSE==maybeexec(cx,glob,rc)) goto failure;
#endif
    
  }

  return cx;

 failure:
  if (cx) JS_DestroyContext(cx);
  return 0;
}

char *evalexpr=0;

static void *startthread(void *v) {
  char *buf=0;
  int fd=0;
  JSContext *cx;
  jsval rval;


#ifdef _WIN32

	while (optind < argc) {
		if (strcmp(argv[optind],"/r")==0 && argc>optind+1) {
			rcfile=argv[++optind];
			optind++;
		} else if (strcmp(argv[optind],"/n")==0) {
			norc=1;
			optind++;
		} else if (strcmp(argv[optind],"/?")==0) {
			printhelp();
			exit(0);
		} else if (strcmp(argv[optind],"/e")==0 && argc>optind+1) {
			evalexpr=argv[++optind];
			optind++;
		} else break;
	}

#else



  for (;;) {
    static struct option long_options[] = {
      {"rcfile", 1, 0, 'r'},
      {"norc", 0, 0, 'n'},
      {"help", 0, 0, 'h'},
      {"eval", 1, 0, 'e'},
      {0, 0, 0, 0}
    };
    
    int c = getopt_long (argc, argv, "r:nhe:",
			  long_options,0);
    
    switch(c) {
    case 'e':
      evalexpr=optarg;
      break;
    case 'r':
      rcfile=optarg;
      break;
    case 'n':
      norc=1;
      break;
    case 'h':
      printhelp();
      exit(0);
      break;
    }

    if (c == -1)
      break;
  }

#endif

#ifdef _WIN32
# ifdef UNICODE
  if (evalexpr) {
	  if (SetEnvironmentVariable(L"JSEXT_INTERACTIVE",L"cmdline")==0)
		  fprintf(stderr,"Setenv error %d\n",GetLastError());
  }
# else
  if (evalexpr) {
	  if (SetEnvironmentVariable("JSEXT_INTERACTIVE","cmdline")==0)
		  fprintf(stderr,"Setenv error %d\n",GetLastError());
  }
# endif
#else
  if (evalexpr) setenv("JSEXT_INTERACTIVE","cmdline",1);
#endif

  cx=JSEXT_InitContext(rt);
  if (!cx) goto failure;

  glob=JS_GetGlobalObject(cx);
  rval=JSVAL_ZERO;
    
  if (optind<argc) {
    struct stat S;
    int shebanglen=0;
    JSScript *script;
    
#ifdef _WIN32
    fd=open(argv[optind],O_RDONLY | O_BINARY);
#else
    fd=open(argv[optind],O_RDONLY);
#endif
    if (fd==-1) {
      JSCC_ReportException(cx, "Error opening %s",argv[optind]);
      goto failure;
    }
    fstat(fd, &S);
    buf=malloc(S.st_size);
    if (!buf) {
      JSCC_ReportException(cx, "Out of memory for %s", argv[optind]);
      goto failure;
    }
    if (read(fd, buf, S.st_size)!=S.st_size) {
      JSCC_ReportException(cx, "Error loading %s",argv[optind]);
      goto failure;
    }
    
    if (S.st_size>1 && buf[0]=='#' && buf[1]=='!') {
      // shebang
      while (buf[shebanglen]!='\n' &&
	     buf[shebanglen]!='\r') shebanglen++;
      shebanglen++;
      if (shebanglen<S.st_size &&
	  (buf[shebanglen]=='\n' ||
	   buf[shebanglen]=='\r') &&
	  buf[shebanglen]!=buf[shebanglen-1]) shebanglen++;
    }

    script=JS_CompileScript(cx, glob, buf+shebanglen, S.st_size-shebanglen, argv[optind], shebanglen?2:1);
    if (!script) goto failure;
    free(buf);
    buf=0;
    
    close(fd);
    fd=0;
    
    if (!JS_ExecuteScript(cx, glob, script, &rval)) goto failure;
    JS_DestroyScript(cx, script);
  }

  if (evalexpr) {
    JSScript *script;
//    if (setenv("JSEXT_INTERACTIVE","1",1)) perror("setenv");
    script=JS_CompileScript(cx, glob, evalexpr, strlen(evalexpr), "<cmdline>", 1);
    if (!script) goto failure;
    if (!JS_ExecuteScript(cx, glob, script, &rval)) goto failure;
    JS_DestroyScript(cx, script);
  }
      
  JS_DestroyContext(cx);
  
  return (void *)rval;
  
 failure:
  if (cx) JS_DestroyContext(cx);
  if (fd) close(fd);
  if (buf) free(buf);
  
  return (void *)JSVAL_NULL;
}

#ifdef _WINDOWS

int
main(int _argc, char **_argv, char **_envp);

int WINAPI WinMain(HINSTANCE instance, HINSTANCE prev_instance, char* command_line, int show_command)
{
    int    argc=0;
	int		argcap=16;
    char** argv;
	char **origargv;
    int    result;
	char *cs;
    char filename[_MAX_PATH];

    // to global
	hInstance=instance;
	hPrevInstance=prev_instance;
	lpCmdLine=command_line;
	nCmdShow=show_command;

    // tokenize the arguments

	cs= command_line;
	argcap=128;
	origargv=argv=(char **)malloc(sizeof(char *)*argcap);

	// put the program name into argv[0]

    GetModuleFileNameA(NULL, filename, _MAX_PATH);
    argv[0] = strdup(filename);
	argc=1;

	for(;;) {
		while (*cs==' ') ++cs; // eat spaces between arguments
		if (!*cs) break;

		if (argc==argcap) {
			argcap*=2;
			argv=(char **)realloc(argv, sizeof(char *)*argcap);
		}
		if (*cs=='"') {
			char *out;
			cs++;
			argv[argc++]=cs;
			out=cs;

			for(;;) {
				if (!*cs) break;
				if (cs[0]=='"' && cs[1]=='"' && cs[2]=='"') cs+=2;
				else if (cs[0]=='"') break;
				*(out++)=*(cs++);
			}
			*out=0;
		} else {
			argv[argc++]=cs;
			if (strchr(cs,' ')) {
				cs=strchr(cs,' ');
				*(cs++)=0;
			} else {
				break;
			}
		}
	};
	argv=origargv;
    result = main(argc, argv, 0);
    
    free(argv);
    return result;
}
#endif

int
main(int _argc, char **_argv, char **_envp)
{
//  pthread_t runner;

  jsval rval;

  argv=_argv;
  envp=_envp;
  argc=_argc;

  rt=JS_NewRuntime(64000000);
  if (!rt) goto failure;


  rval=(jsval)startthread(0);
  /*
  // Threadsafe spidermonkey only works in a thread
  if (pthread_create(&runner, NULL, startthread, 0)) {
    perror("Creating thread");
    exit(1);
    }

  jsval rval;
  if ((jsval)pthread_join(runner, (void **)&rval)) {
    perror("Ending runner thread");
  }
  */

  return 0;
  //  if (JSVAL_IS_INT(rval)) return JSVAL_TO_INT(rval);

 failure:
  return 255;
}

